// ============================================================================
// $Id: lpc_protocol.cc,v 1.1 1996/04/22 18:25:24 questor Exp $
//
// This is part of Realm Of Magic
//
// Usage: Class that translates incoming messages to the lpc protocol
//
// All rights reserved. See license.doc for complete information.
//
// Copyright (C) 1993-1996 by the Hackerz of the Realm Of Magic
// Realm Of Magic is based on CircleMUD, Copyright (C) 1993.
// CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.
// Original intermud_client.c (C) 1995, 96 by Chris Austin (Stryker@Tempus)
// Original lpc protocol (C) by Mark Lewis (Nostradamus@Zebedee)
//                           by J.W. Armstrong (Alvin@Sushi)
// ============================================================================

#include <string>
#include <list>
#include <fstream.h>
#include <stdio.h>
#include <strstream.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "net.hh"
#include "log.hh"
#include "tokenizer.hh"
#include "mud.hh"
#include "lpc_protocol.hh"
#include "utils.hh"

extern _log interlog;
extern _mud localmud;
extern _net network;
extern int lpc_id_number;

void _lpc::Startup()
{
  dns_record dummy;
  ifstream mudlist(MUDLIST_LPC_PATH);
  char buffer[8192];
  tokenizer tokens;

  if (!mudlist) {
    interlog << "SYSERR: no lpc mudlist, exiting!" << logit;
    exit(1);
  }

  dummy.kb_in = 0;
  dummy.kb_out = 0;
  dummy.msg_in = 0;
  dummy.msg_out = 0;
  dummy.muted = 0;
  dummy.last_connected = "No connection yet";

  while (!mudlist.eof()) {
    mudlist.getline(buffer, 8192, '\n');
    tokens.tokenize(buffer, "|");
    if (tokens[0]) {
      dummy.ip_address = inet_addr(tokens[2]);
      dummy.port = htons(atoi(tokens[3]));
      dummy.mud_name = tokens[0];
      dummy.mud_port = tokens[1];
      dummy.mud_version = tokens[5];
      dummy.intermud_version = tokens[6];
      dummy.mud_type = atoi(tokens[4]);
      dummy.services = 0;
      decode_services(tokens[7], dummy);
      dns_table.push_back(dummy);
    }
    tokens.clear();
  }
  mudlist.close();
  return;
}

void _lpc::SendPings()
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position) {
    network.SetAddr((*position).ip_address);
    network.SetPort((*position).port);
    network << "REQ:ping|UDP:" << INTERMUD_PORT << "|ID:"
            << ++lpc_id_number << "|SND:Server|NAME:" << MUDNAME 
            << sendit;

    lpc_data.id_nr = lpc_id_number;
    lpc_data.udp_port = (*position).port;
    lpc_data.mud_name = (*position).mud_name;
    lpc_data.channel = "";
    lpc_data.request = "";
    lpc_data.sender = "";
    lpc_data.recipent = "";
    lpc_data.reply_type = "ping";
    reply_table.push_front(lpc_data);
  }
  return;
}

void _lpc::message2lpc(string &mess)
{
  lpc_data.id_nr = 0;
  lpc_data.udp_port = 0;
  lpc_data.mud_name = "";
  lpc_data.channel = "";
  lpc_data.request = "";
  lpc_data.sender = "";
  lpc_data.recipent = "";
  lpc_data.data = "";
  lpc_data.reply_type = "";

  string::size_type pos1 = 0;
  string::size_type pos2 = 0;
  int deadlock = 0;
  while (1) {
    deadlock++;
    if (deadlock == 1500) {
      interlog << "DEADLOCK!! EXITING!!" << logit;
      interlog << "Mess string was: " << mess << logit;
      exit(1);
    }
    if (!case_n_compare(mess.substr(pos1), "UDP:", 4)) {
      pos2 = mess.find('|', pos1);
      if (pos2 == string::npos)
        return;
      pos1 = pos2 + 1;
      continue;
    }
    if (!case_n_compare(mess.substr(pos1), "ID:", 3)) {
      pos2 = mess.find('|', pos1);
      if (pos2 == string::npos) {
        lpc_data.id_nr = atoi(mess.substr(pos1 + 3).c_str());
        return;
      } else
        lpc_data.id_nr = atoi(mess.substr(pos1 + 3,
                              pos2 - (pos1 + 3)).c_str());
      pos1 = pos2 + 1;
      continue;
    }
    if (!case_n_compare(mess.substr(pos1), "CHANNEL:", 8)) {
      pos2 = mess.find('|', pos1);
      if (pos2 == string::npos) {
        lpc_data.channel = mess.substr(pos1 + 8);
        return;
      } else
        lpc_data.channel = mess.substr(pos1 + 8,
                                   pos2 - (pos1 + 8));
      pos1 = pos2 + 1;
      continue;
    }
    if (!case_n_compare(mess.substr(pos1), "RCPNT:", 6)) {
      pos2 = mess.find('|', pos1);
      if (pos2 == string::npos) {
        lpc_data.recipent = mess.substr(pos1 + 6);
        return;
      } else
        lpc_data.recipent = mess.substr(pos1 + 6,
                                    pos2 - (pos1 + 6));
      pos1 = pos2 + 1;
      continue;
    }
    if (!case_n_compare(mess.substr(pos1), "REQ:", 4)) {
      pos2 = mess.find('|', pos1);
      if (pos2 == string::npos) {
        lpc_data.request = mess.substr(pos1 + 4);
        return;
      } else
        lpc_data.request = mess.substr(pos1 + 4,
                                   pos2 - (pos1 + 4));
      pos1 = pos2 + 1;
      continue;
    }
    if (!case_n_compare(mess.substr(pos1), "SND:", 4)) {
      pos2 = mess.find('|', pos1);
      if (pos2 == string::npos) {
        lpc_data.sender = mess.substr(pos1 + 4);
        return;
      } else
        lpc_data.sender = mess.substr(pos1 + 4,
                                  pos2 - (pos1 + 4));
      pos1 = pos2 + 1;
      continue;
    }
    if (!case_n_compare(mess.substr(pos1), "DATA:", 5)) {
      lpc_data.data = mess.substr(pos1 + 5);
      return;
    }
    if (!case_n_compare(mess.substr(pos1), "NAME:", 5)) {
      pos2 = mess.find('|', pos1);
      if (pos2 == string::npos) {
        lpc_data.mud_name = mess.substr(pos1 + 5);
        return;
      } else
        lpc_data.mud_name = mess.substr(pos1 + 5,
                                    pos2 - (pos1 + 5));
      pos1 = pos2 + 1;
      continue;
    }
    pos2 = mess.find('|', pos1);
    if (pos2 == string::npos)
      return;
    pos1 = pos2 + 1;
  }
  return;
}

bool _lpc::IncomingMessage(string &mess, sockaddr_in &remote_address)
{
  // Checking if this is really a lpc message
  if (case_n_compare(mess, "PKT:", 4) &&
      case_n_compare(mess, "UDP:", 4) &&
      case_n_compare(mess, "ID:", 3) &&
      case_n_compare(mess, "CHANNEL:", 8) &&
      case_n_compare(mess, "RCPNT:", 6) &&
      case_n_compare(mess, "REQ:", 4) &&
      case_n_compare(mess, "SND:", 4) &&
      case_n_compare(mess, "DATA:", 5) &&
      case_n_compare(mess, "NAME:", 5))
    return false;
  
  if (!case_n_compare(mess, "PKT:", 4)) {
    HandleSplitPackets(mess, remote_address);
    return true;
  }
  message2lpc(mess);
  if (!case_compare(lpc_data.request, "reply"))
    IncomingReply(mess, remote_address);
  else if (!case_compare(lpc_data.channel, "intercode"))
    IncomingInterwiz();
  else if (!case_compare(lpc_data.request, "tell"))
    IncomingIntertell(remote_address);
  else if (!case_compare(lpc_data.request, "ping"))
    IncomingPing(remote_address);
  else if (!case_compare(lpc_data.request, "who"))
    IncomingWhoREQ();
  else {
    if (lpc_data.channel.empty()) {
      interlog << "RNET: [ UNKNOWN] " << mess << " - "
               << inet_ntoa(remote_address.sin_addr) << logit;
      if (!lpc_data.sender.empty()) {
        network.SetAddr(remote_address.sin_addr.s_addr);
        network.SetPort(remote_address.sin_port);
        network << "ID:" << lpc_data.id_nr << "|RCPNT:" << lpc_data.sender
                << "|REQ:reply|UDP:" << INTERMUD_PORT << "|NAME:"
                << MUDNAME << "|DATA:Invalid request @" << MUDNAME
                << ": " << lpc_data.request << "\n" << sendit;
      }
      return false;
    }
  }
  return true;
}

void _lpc::IncomingReply(string &mess, sockaddr_in &remote_address)
{
  bool found = false;
  int idx;

  list<lpc_info_record>::iterator lipos;
  for (lipos = reply_table.begin(); lipos != reply_table.end();
      ++lipos)
    if ((*lipos).id_nr == lpc_data.id_nr) {
      found = true;
      break;
    }
  if (!found) {
    interlog << "SYSERR: Received packet with unknown id-nr " << lpc_data.id_nr
             << " from " << inet_ntoa(remote_address.sin_addr) << logit;
    return;
  }
  if (!case_compare((*lipos).reply_type, "ping")) {
    if (!case_n_compare((*lipos).mud_name, lpc_data.data, 
        (*lipos).mud_name.size())) {
       list<dns_record>::iterator position;
       for (position = dns_table.begin(); position != dns_table.end();
           ++position)
         if ((*position).ip_address == remote_address.sin_addr.s_addr) {
           update_timestamp(*position);
           reply_table.erase(lipos);
           return;
         }
       return;
    } else {
      interlog << "SYSERR: Unknown packet with id number of ping-request "
                  "received from "
               << inet_ntoa(remote_address.sin_addr) << logit;
      return;
    }
  } else if (!case_compare((*lipos).reply_type, "tell")) {
    if (!case_compare((*lipos).mud_name, lpc_data.mud_name)) {
      list<dns_record>::iterator position;
      for (position = dns_table.begin(); position != dns_table.end();
          ++position)
        if ((*position).ip_address == remote_address.sin_addr.s_addr) {
          update_timestamp(*position);
          reply_table.erase(lipos);
          localmud << "inc-tell-nr#@#Server#@#RealmOfMagic#@#"
                   << lpc_data.recipent 
                   << "#@#Your message has been delivered!" << sendit;
          return;
        }
      return;
    } else {
      interlog << "SYSERR: Unknown packet with id number of tell-ack "
                  "received from "
               << inet_ntoa(remote_address.sin_addr)
               << ": " << mess << logit;
      return;
    }
  } else if (!case_compare((*lipos).reply_type, "who")) {
    if (!case_compare((*lipos).mud_name, lpc_data.mud_name)) {
      list<dns_record>::iterator position;
      for (position = dns_table.begin(); position != dns_table.end();
          ++position) {
        if (!case_compare((*position).mud_name, (*lipos).mud_name)) {
          update_timestamp(*position);
          localmud << "inc-who-reply#@#" << lpc_data.recipent << "#@#"
                   << lpc_data.mud_name << "#@#" << lpc_data.data << sendit;
          reply_table.erase(lipos);
          return;
        }
      }
    }
  } else {
    interlog << "SYSERR: Unexpected reply packet received from "
             << inet_ntoa(remote_address.sin_addr)
             << ": " << lpc_data.data << logit;
    return;
  }

}

void _lpc::IncomingInterwiz()
{
  localmud << "inc-interwiz#@#" << lpc_data.sender << "#@#"
           << lpc_data.mud_name << "#@#" << lpc_data.data << sendit;
  return;
}

void _lpc::IncomingIntertell(sockaddr_in &remote_address)
{
  localmud << "inc-tell-nr#@#" << lpc_data.sender << "#@#" << lpc_data.mud_name
           << "#@#" << lpc_data.recipent << "#@#" << lpc_data.data << sendit;
  network.SetAddr(remote_address.sin_addr.s_addr);
  network.SetPort(remote_address.sin_port);
  network << "ID:" << lpc_data.id_nr << "|" << "REQ:reply|RCPNT:"
          << lpc_data.sender << "|NAME:" << MUDNAME << "|UDP:"
          << INTERMUD_PORT << "|DATA:Server@" << MUDNAME
          << " tells you, 'Message received!'";
  network << sendit;
  return;
}

void _lpc::IncomingPing(sockaddr_in &remote_address)
{
  network.SetAddr(remote_address.sin_addr.s_addr);
  network.SetPort(remote_address.sin_port);
  network << "ID:" << lpc_data.id_nr << "|NAME:" << MUDNAME << "|UDP:"
          << INTERMUD_PORT;
  if (!lpc_data.sender.empty())
    network << "|RCPNT:" << lpc_data.sender;
  network << "|REQ:reply|DATA:" << MUDNAME << " is alive.\n" << sendit;
  return;
}

void _lpc::IncomingWhoREQ()
{
  strstream tmp_num;
  lpc_info_record reply_rec;
  reply_rec.id_nr = ++lpc_id_number;
  reply_rec.mud_name = lpc_data.mud_name;
  reply_rec.sender = lpc_data.sender;
  tmp_num << lpc_data.id_nr << ends;
  reply_rec.data = tmp_num.str();
  reply_table.push_front(reply_rec);
  localmud << "inc-who-query#@#" << lpc_data.sender << "#@#"
           << lpc_data.mud_name << sendit;
  return;
}

void _lpc::HandleSplitPackets(string &mess, sockaddr_in &remote_address)
{
  string::size_type str_pos = mess.find('|');
  string id_part = mess.substr(0, str_pos);
  ++str_pos;
  string packet_part = mess.substr(str_pos);
  string tmp;
  
  tokenizer pkt_id(id_part, ":");
  tokenizer id_num(pkt_id[3], "/");

  int max_num = atoi(id_num[1]);
  int cur_num = atoi(id_num[0]);

  lpc_data.id_nr = atoi(pkt_id[2]);

  if (cur_num == max_num) {
    lpc_data.data = packet_part;
    lpc_data.udp_port = 0;
    lpc_data.mud_name = pkt_id[1];
    lpc_data.channel = "";
    lpc_data.request = "";
    lpc_data.sender = "";
    lpc_data.recipent = "";
    packet_table.push_front(lpc_data);
    return;
  } else if (cur_num == 1) {
    list<lpc_info_record>::iterator lipos;
    for (lipos = packet_table.begin(); lipos != packet_table.end();
        ++lipos)
      if ((*lipos).id_nr == lpc_data.id_nr) {
        message2lpc(packet_part);
        lpc_data.data += (*lipos).data;
        if (!case_compare(lpc_data.request, "reply"))
          IncomingReply(lpc_data.data, remote_address);
        else if (!case_compare(lpc_data.channel, "intercode"))
          IncomingInterwiz();
        else if (!case_compare(lpc_data.request, "tell"))
          IncomingIntertell(remote_address);
        else if (!case_compare(lpc_data.request, "ping"))
          IncomingPing(remote_address);
        else
          interlog << "INFO: Invalid split packed received from ("
                   << inet_ntoa(remote_address.sin_addr) << ")" << mess
                   << logit;
        packet_table.erase(lipos);
        return;
      }
  } else {
    list<lpc_info_record>::iterator lipos;
    for (lipos = packet_table.begin(); lipos != packet_table.end();
        ++lipos)
      if ((*lipos).id_nr == lpc_data.id_nr) {
        tmp = (*lipos).data;
        (*lipos).data = packet_part;
        (*lipos).data += tmp;
        return;
      }
    return;
  }
}
 
bool _lpc::OutgoingIntertell(string &recipent, string &remote_mud,
                             string &player, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
       ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      lpc_data.id_nr = ++lpc_id_number;
      lpc_data.udp_port = (*position).port;
      lpc_data.mud_name = (*position).mud_name;
      lpc_data.channel = "";
      lpc_data.request = "";
      lpc_data.sender = "";
      lpc_data.recipent = "";
      lpc_data.data = "";
      lpc_data.reply_type = "tell";
      reply_table.push_front(lpc_data);
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "ID:" << lpc_id_number << "|NAME:" << MUDNAME
              << "|REQ:tell|RCPNT:" << recipent << "|SND:"
              << player << "|UDP:" << INTERMUD_PORT << "|DATA:"
              << data;
      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;
      return true;
    }
  return false;
}

bool _lpc::OutgoingIntertellNR(string &recipent, string &remote_mud,
                               string &player, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
       ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      lpc_data.id_nr = ++lpc_id_number;
      lpc_data.udp_port = (*position).port;
      lpc_data.mud_name = (*position).mud_name;
      lpc_data.channel = "";
      lpc_data.request = "";
      lpc_data.sender = "";
      lpc_data.recipent = "";
      lpc_data.data = "";
      lpc_data.reply_type = "tell";
      
      if (!case_compare(player, "Server"))
        return true;

      reply_table.push_front(lpc_data);
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "ID:" << lpc_id_number << "|NAME:" << MUDNAME
              << "|REQ:tell|RCPNT:" << recipent << "|SND:"
              << player << "|UDP:" << INTERMUD_PORT << "|DATA:"
              << data;
      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;
      return true;
    }
  return false;
}

bool _lpc::OutgoingInterwho(string &recipent, string &remote_mud, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      list<lpc_info_record>::iterator lpr_pos;
      bool found = false;
      for (lpr_pos = reply_table.begin(); lpr_pos != reply_table.end();
          ++lpr_pos) {
        if ((!case_compare((*lpr_pos).mud_name, (*position).mud_name)) &&
            (!case_compare((*lpr_pos).sender, recipent))) {
          found = true;
          break;
        }
      }
      if (found == false)
        return false;


// HHH   HHH AAAAAAAAA CCCCCCCCC KK  KK
// HHH   HHH AAA   AAA CC        KK  KK
// HHHHHHHHH AAAAAAAAA CC        KKKK
// HHH   HHH AAA   AAA CC        KK  KK
// HHH   HHH AAA   AAA CCCCCCCCC KK  KK

      string header = "RCPNT:"; header += recipent;
      header += "|ID:"; header += (*lpr_pos).data;
      header += "|NAME:"; header += MUDNAME; header += "|REQ:reply|UDP:";
      header += INTERMUD_PORT; header += "|DATA:";
      lpc_id_number++;
      string tmp_str;
      if (header.size() + data.size() > 1024) {

        tokenizer who_data(data, "\n");

        int max_paks = who_data.size() / 10;
        if (who_data.size() % 10 != 0)
          max_paks++;

        int curr_pak;
        int data_idx = who_data.size() - 1;
        int y;
        string pak_data = "";
        for (curr_pak = max_paks; curr_pak > 1; --curr_pak) {
          network << "PKT:" << MUDNAME << ":" << lpc_id_number
                  << ":" << curr_pak << "/" << max_paks
                  << "|";
          pak_data = "";
          for (y = 0; y < 10 && data_idx >= 0; ++y) {
            tmp_str = who_data[data_idx];
            tmp_str += "\n";
            tmp_str += pak_data;
            pak_data = tmp_str;
            --data_idx;
          }
          network << pak_data;
          pak_data = "";
          network << sendit;
          network.SetAddr((*position).ip_address);
          network.SetPort((*position).port);
        }
        for (y = 0; y < 10 && data_idx >= 0; ++y) {
          tmp_str = who_data[data_idx];
          tmp_str += "\n";
          tmp_str += pak_data;
          pak_data = tmp_str;
          --data_idx;
        }
        network << "PKT:" << MUDNAME << ":" << lpc_id_number
                << ":" << curr_pak << "/" << max_paks
                << "|" << header << pak_data << sendit;
        return true;      
      } else {
        network << "RCPNT:" << recipent << "|ID:" << (*lpr_pos).data
                << "|NAME:" << MUDNAME << "|REQ:reply|UDP:"
                << INTERMUD_PORT << "|DATA:" << data << sendit;
        return true;
      }
      reply_table.erase(lpr_pos);
    }
  return false;
}

bool _lpc::OutgoingInterwhoREQ(string &remote_mud, string &player)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
      ++position)
    if (!case_compare((*position).mud_name, remote_mud)) {
      network.SetAddr((*position).ip_address);
      network.SetPort((*position).port);
      network << "ID:" << ++lpc_id_number << "|UDP:" << INTERMUD_PORT
              << "|NAME:" << MUDNAME << "|SND:" << player
              << "|REQ:who";

      update_statistics((*position).mud_name, 0, network.OutBuffer().size());
      network << sendit;

      lpc_data.id_nr = lpc_id_number;
      lpc_data.udp_port = (*position).port;
      lpc_data.mud_name = (*position).mud_name;
      lpc_data.channel = "";
      lpc_data.request = "";
      lpc_data.sender = player;
      lpc_data.recipent = "";
      lpc_data.data = "";
      lpc_data.reply_type = "who";
      reply_table.push_front(lpc_data);
      return true;
    }
  return false;
   
}

void _lpc::OutgoingInterwiz(string &player, string &data)
{
  list<dns_record>::iterator position;
  for (position = dns_table.begin(); position != dns_table.end();
       ++position) {
    network.SetAddr((*position).ip_address);
    network.SetPort((*position).port);
    network << "CHANNEL:intercode|REQ:channel|SND:" << player << "|NAME:"
            << MUDNAME << "|UDP:" << INTERMUD_PORT << "|DATA:" << data;
    update_statistics((*position).mud_name, 0, network.OutBuffer().size());
    network << sendit;
  }
  return;
}

